package org.xbib.jdbc.query;

import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Primary class for accessing a relational database.
 */
public class DatabaseImpl implements Database {
    
    private static final Logger logger = Logger.getLogger(Database.class.getName());
    
    private final Connection connection;

    private final Options options;

    public DatabaseImpl(Flavor flavor) {
        this(new OptionsDefault(flavor));
    }

    public DatabaseImpl(Options options) {
        this(null, options);
    }

    public DatabaseImpl(Connection connection, Options options) {
        this.connection = connection;
        this.options = options;
    }

    @Override
    public DatabaseImpl get() {
        return this;
    }

    @Override
    public SqlInsert toInsert(String sql) {
        return new SqlInsertImpl(connection, sql, options);
    }

    @Override
    public SqlInsert toInsert(Sql sql) {
        return new SqlInsertImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public SqlSelect toSelect(String sql) {
        return new SqlSelectImpl(connection, sql, options);
    }

    @Override
    public SqlSelect toSelect(Sql sql) {
        return new SqlSelectImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public SqlUpdate toUpdate(String sql) {
        return new SqlUpdateImpl(connection, sql, options);
    }

    @Override
    public SqlUpdate toUpdate(Sql sql) {
        return new SqlUpdateImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public SqlUpdate toDelete(String sql) {
        return new SqlUpdateImpl(connection, sql, options);
    }

    @Override
    public SqlUpdate toDelete(Sql sql) {
        return new SqlUpdateImpl(connection, sql.sql(), options).apply(sql);
    }

    @Override
    public Ddl ddl(String sql) {
        return new DdlImpl(connection, sql, options);
    }

    @Override
    public Long nextSequenceValue(String sequenceName) {
        return toSelect(flavor().sequenceSelectNextVal(sequenceName)).queryLongOrNull();
    }

    @Override
    public void commitNow() {
        if (options.ignoreTransactionControl()) {
            logger.fine("Ignoring call to commitNow()");
            return;
        }
        if (!options.allowTransactionControl()) {
            throw new DatabaseException("Calls to commitNow() are not allowed");
        }
        try {
            connection.commit();
        } catch (Exception e) {
            throw new DatabaseException("Unable to commit transaction", e);
        }
    }

    @Override
    public void rollbackNow() {
        if (options.ignoreTransactionControl()) {
            logger.fine("Ignoring call to rollbackNow()");
            return;
        }
        if (!options.allowTransactionControl()) {
            throw new DatabaseException("Calls to rollbackNow() are not allowed");
        }
        try {
            connection.rollback();
        } catch (Exception e) {
            throw new DatabaseException("Unable to rollback transaction", e);
        }
    }

    @Override
    public Connection underlyingConnection() {
        if (!options.allowConnectionAccess()) {
            throw new DatabaseException("Calls to underlyingConnection() are not allowed");
        }
        return connection;
    }

    @Override
    public Options options() {
        return options;
    }
    
    @Override
    public Flavor flavor() {
        return options.flavor();
    }

    @Override
    public int jdbcMajorVersion() {
        try {
            return connection.getMetaData().getJDBCMajorVersion();
        } catch (SQLException e) {
            throw new DatabaseException(e);
        }
    }

    @Override
    public int jdbcMinorVersion() {
        try {
            return connection.getMetaData().getJDBCMinorVersion();
        } catch (SQLException e) {
            throw new DatabaseException(e);
        }
    }

    @Override
    public When when() {
        return new When(options.flavor());
    }

    @Override
    public void dropSequenceQuietly(String sequenceName) {
        ddl(flavor().sequenceDrop(sequenceName)).executeQuietly();
    }

    @Override
    public void dropTableQuietly(String tableName) {
        ddl(flavor().tableDrop(tableName)).executeQuietly();
    }

    @Override
    public boolean tableExists(String tableName) throws DatabaseException {
        try {
            return tableExists(tableName, connection.getSchema());
        } catch (SQLException e) {
            throw new DatabaseException("Unable to getSchema()", e);
        }
    }

    @Override
    public boolean tableExists(String tableName, String schemaName) throws DatabaseException {
        if (tableName != null && connection != null) {
            try {
                DatabaseMetaData metadata = connection.getMetaData();
                String normalizedTable = flavor().normalizeTableName(tableName);
                ResultSet resultSet = metadata.getTables(connection.getCatalog(), schemaName, normalizedTable, new String[]{"TABLE", "VIEW"});
                while (resultSet.next()) {
                    if (normalizedTable.equals(resultSet.getString("TABLE_NAME"))) {
                        return true;
                    }
                }
            } catch (SQLException exc) {
                throw new DatabaseException("Unable to look up table " + tableName
                        + " in schema  " + schemaName + " : " + exc.getMessage(),
                        exc);
            }
        }
        return false;
    }

    @Override
    public Map<String, Integer> getColumnSizesOfTable(String tableName) {
        Map<String, Integer> map = new HashMap<>();
        if (tableName != null && connection != null) {
            try {
                DatabaseMetaData metaData = connection.getMetaData();
                String normalizedTable = flavor().normalizeTableName(tableName);
                ResultSet resultSet = metaData.getColumns(null, null, normalizedTable, "%");
                ResultSetMetaData rsmd = resultSet.getMetaData();
                int cols = rsmd.getColumnCount();
                while (resultSet.next()) {
                    String name = "";
                    int size = 0;
                    for (int i = 1; i <= cols; i++) {
                        String label = rsmd.getColumnName(i);
                        if ("COLUMN_NAME".equalsIgnoreCase(label)) {
                            name = resultSet.getString(i);
                        }
                        if ("COLUMN_SIZE".equalsIgnoreCase(label)) {
                            size = resultSet.getBigDecimal(i).intValue();
                        }
                    }
                    map.put(name, size);
                }
                resultSet.close();
            } catch (SQLException exc) {
                throw new DatabaseException("Unable to look up table " + tableName + " : " + exc.getMessage(), exc);
            }
        }
        return map;
    }

    @Override
    public void assertTimeSynchronized(long millisToWarn, long millisToError) {
        toSelect("select ?" + flavor().fromAny())
                .argLocalDateTimeNowPerDb()
                .queryFirstOrNull(r -> {
                    LocalDateTime appDate = LocalDateTime.now();
                    LocalDateTime dbDate = r.getLocalDateTimeOrNull();
                    if (dbDate == null) {
                        throw new DatabaseException("Expecting a date in the result");
                    }
                    Duration duration = Duration.between(appDate, dbDate).abs();
                    if (duration.getSeconds() > 3600) {
                        throw new DatabaseException("JDBC and database time are over an hour apart (check your timezones) app: "
                                + appDate + " db: "
                                + dbDate);
                    }
                    if (duration.getSeconds() * 1000 > millisToError) {
                        throw new DatabaseException("JDBC and database time over " + millisToError + " millis apart (check your clocks) app: "
                                + appDate + " db: "
                                + dbDate);
                    }
                    if (duration.getSeconds() * 1000 > millisToWarn) {
                        logger.warning("JDBC and database time are over " + millisToWarn + " millis apart (check your clocks) app: "
                                + appDate + " db: "
                                + dbDate);
                    }
                    return null;
                });
    }

    @Override
    public void assertTimeSynchronized() {
        assertTimeSynchronized(10000, 30000);
    }

    @Override
    public Table getPagedRows(Table table) {
        if (table == null) {
            return table;
        }
        Map<String, Object> params = table.getParams() != null ?
                new HashMap<>(table.getParams()) : new HashMap<>();
        if (table.getOffset() != null && table.getSize() != null) {
            params.put("offset", table.getOffset());
            params.put("limit", table.getSize());
        }
        String where = table.getWhereClause() != null ? table.getWhereClause() : "";
        String groupby = table.getGroupByClause() != null ? table.getGroupByClause() : "";
        String orderby = !table.getSort().isEmpty() ? "order by " + table.getSort() : "";
        String statement = table.getStatement() + " " + where + " " + groupby + " " + orderby;
        if (table.getOffset() != null && table.getSize() != null) {
            statement = statement + " offset :offset rows fetch next :limit rows only";
        }
        return getUnlimitedRows(statement, params);
    }

    @Override
    public long getRowsCount(Table table) {
        Map<String, Object> params = new HashMap<>(table.getParams());
        String statement = table.getStatement() + " " + table.getWhereClause();
        return countRows(statement, params);
    }

    @Override
    public long countRows(String statement, Map<String, Object> params) {
        String countStatament = "select count(*) as \"cnt\" from (" + statement + ")";
        Table table = getSingleRow(countStatament, params);
        if (!table.isEmpty()) {
            BigDecimal bigDecimal = table.getValue(0,"cnt");
            return bigDecimal.longValue();
        } else {
            return -1L;
        }
    }

    @Override
    public Table getSingleRow(String statement, Map<String, Object> params) {
        return getLimitedRows(statement, params, 1, options().fetchSize(), options().timeoutSeconds());
    }

    @Override
    public Table getUnlimitedRows(String statement, Map<String, Object> params) {
        return getLimitedRows(statement, params, 0, options().fetchSize(), options().timeoutSeconds());
    }

    @Override
    public Table getLimitedRows(String statement, Map<String, Object> params,
                                int limit, int fetchSize, int timeoutSeconds) {
        SqlSelect sql = toSelect(statement).fetchSize(fetchSize).withTimeoutSeconds(timeoutSeconds);
        selectParams(sql, params);
        Table table = new Table();
        sql.query(rows -> {
            ResultSetMetaData md = rows.getMetadata();
            List<Object> columnNames = new ArrayList<>();
            List<Object> classNames = new ArrayList<>();
            for (int i = 1; i <= md.getColumnCount(); i++) {
                columnNames.add(md.getColumnName(i).toLowerCase(Locale.ROOT));
                classNames.add(md.getColumnClassName(i));
            }
            table.add(columnNames);
            table.add(classNames);
            int i = 0;
            while (rows.next() && (limit <= 0 || i++ < limit)) {
                table.add(getRow(rows, classNames));
            }
            table.setTotal(rows.rowCount());
            return true;
        });
        return table;
    }

    @Override
    public void insert(String statement, Map<String, Object> params) {
        SqlInsert sql = toInsert(statement);
        insertParams(sql, params);
        sql.insert(1);
    }

    @Override
    public void insert(String statement, List<Map<String, Object>> params) {
        SqlInsert sqlInsert = toInsert(statement);
        for (Map<String, Object> param : params) {
            insertParams(sqlInsert, param);
            sqlInsert.batch();
        }
        sqlInsert.insertBatch();
    }

    @Override
    public void update(String statement, Map<String, Object> params) {
        SqlUpdate sql = toUpdate(statement);
        updateParams(sql, params);
        sql.update();
    }

    @Override
    public void update(String statement, List<Map<String, Object>> params) {
        SqlUpdate sqlUpdate = toUpdate(statement);
        for (Map<String, Object> param : params) {
            updateParams(sqlUpdate, param);
            sqlUpdate.update();
        }
    }

    @Override
    public void upsert(String insertStatement, String updateStatement, Map<String, Object> params) {
        // try insert then update if error
        try {
            SqlInsert sql = toInsert(insertStatement);
            insertParams(sql, params);
            sql.insert(1);
        } catch (Exception e) {
            logger.log(Level.WARNING, e.getMessage(), e);
            SqlUpdate sql = toUpdate(updateStatement);
            updateParams(sql, params);
            sql.update();
        }
    }

    @Override
    public void delete(String statement, Map<String, Object> params) {
        SqlUpdate sql = toDelete(statement);
        updateParams(sql, params);
        sql.update();
    }

    @Override
    public void delete(String statement, List<Map<String, Object>> params) {
        SqlUpdate sqlUpdate = toDelete(statement);
        for (Map<String, Object> param : params) {
            updateParams(sqlUpdate, param);
            sqlUpdate.update();
        }
    }

    private void selectParams(SqlSelect sql, Map<String, Object> params) {
        if (params == null) {
            return;
        }
        params.forEach((k, v) -> {
            if (v instanceof String) {
                sql.argString(k, (String) v);
            } else if (v instanceof Integer) {
                sql.argInteger(k, (Integer) v);
            } else if (v instanceof Long) {
                sql.argLong(k, (Long) v);
            } else if (v instanceof Boolean) {
                sql.argBoolean(k, (Boolean) v);
            } else if (v instanceof LocalDate) {
                sql.argLocalDate(k, (LocalDate) v);
            } else if (v instanceof LocalDateTime) {
                sql.argLocalDateTime(k, (LocalDateTime) v);
            } else {
                throw new DatabaseException("unknown type for param: " + (v != null ? v.getClass() : "null"));
            }
        });
    }

    private void insertParams(SqlInsert sql, Map<String, Object> params) {
        if (params == null) {
            return;
        }
        params.forEach((k, v) -> {
            if (v instanceof String) {
                sql.argString(k, (String) v);
            } else if (v instanceof Integer) {
                sql.argInteger(k, (Integer) v);
            } else if (v instanceof Long) {
                sql.argLong(k, (Long) v);
            } else if (v instanceof Boolean) {
                sql.argBoolean(k, (Boolean) v);
            } else if (v instanceof LocalDate) {
                sql.argLocalDate(k, (LocalDate) v);
            } else if (v instanceof LocalDateTime) {
                sql.argLocalDateTime(k, (LocalDateTime) v);
            } else {
                throw new DatabaseException("unknown type for param: " + (v != null ? v.getClass() : "null"));
            }
        });
    }

    private void updateParams(SqlUpdate sql, Map<String, Object> params) {
        if (params == null) {
            return;
        }
        params.forEach((k, v) -> {
            if (v instanceof String) {
                sql.argString(k, (String) v);
            } else if (v instanceof Integer) {
                sql.argInteger(k, (Integer) v);
            } else if (v instanceof Long) {
                sql.argLong(k, (Long) v);
            } else if (v instanceof Boolean) {
                sql.argBoolean(k, (Boolean) v);
            } else if (v instanceof LocalDate) {
                sql.argLocalDate(k, (LocalDate) v);
            } else if (v instanceof LocalDateTime) {
                sql.argLocalDateTime(k, (LocalDateTime) v);
            } else {
                throw new DatabaseException("unknown type for param: " + (v != null ? v.getClass() : "null"));
            }
        });
    }

    private List<Object> getRow(Rows rows, List<Object> classNames) {
        List<Object> row = new ArrayList<>();
        for (int i = 0; i < classNames.size(); i++) {
            String className = classNames.get(i).toString();
            switch (className) {
                case "java.lang.String":
                    row.add(rows.getStringOrEmpty(i + 1));
                    break;
                case "java.lang.Integer":
                    row.add(rows.getIntegerOrNull(i + 1));
                    break;
                case "java.lang.Long":
                    row.add(rows.getLongOrNull(i + 1));
                    break;
                case "java.lang.Boolean":
                    row.add(rows.getBooleanOrFalse(i + 1));
                    break;
                case "java.sql.Clob":
                case "oracle.jdbc.OracleClob":
                    row.add(rows.getClobStringOrEmpty(i + 1));
                    break;
                case "java.sql.Date":
                    row.add(rows.getLocalDateOrNull(i + 1));
                    break;
                case "java.sql.Timestamp":
                case "oracle.sql.TIMESTAMP":
                    row.add(rows.getLocalDateTimeOrNull(i + 1));
                    break;
                case "java.math.BigDecimal":
                    row.add(rows.getBigDecimalOrNull(i + 1));
                    break;
                default:
                    throw new DatabaseException("unexpected column class name: " + className);
            }
        }
        return row;
    }

    @Override
    public void writeQueue(String table, String channel, String data, Consumer<Long> consumer) throws SQLException {
        writeNextIntoQueue(connection, table, channel, data, consumer);
    }

    @Override
    public void readQueue(String table, String channel, Consumer<String> consumer) throws SQLException {
        try {
            connection.setAutoCommit(false);
            while (true) {
                try (ResultSet resultSet = readNextFromQueue(connection, table, channel)) {
                    if (resultSet.next()) {
                        Long key = resultSet.getLong("key");
                        try {
                            if (consumer != null) {
                                consumer.accept(resultSet.getString("data"));
                            }
                            succeedInQueue(connection, table, key);
                        } catch (Exception e) {
                            failInQueue(connection, table, key);
                            throw e;
                        }
                    }
                } finally {
                    connection.commit();
                }
            }
        } catch (Exception e) {
            connection.rollback();
            throw e;
        } finally {
            connection.setAutoCommit(true);
        }
    }

    private void writeNextIntoQueue(Connection connection,
                                    String table,
                                    String channel,
                                    String data,
                                    Consumer<Long> consumer) throws SQLException {
        try (PreparedStatement preparedStatement = connection.prepareStatement(flavor().writeNextIntoQueue(table),
                Statement.RETURN_GENERATED_KEYS)) {
            preparedStatement.setString(1, channel);
            preparedStatement.setString(2, data);
            preparedStatement.setTimestamp(3, Timestamp.valueOf(LocalDateTime.now()));
            int rows = preparedStatement.executeUpdate();
            if (rows > 0) {
                try (ResultSet resultSet = preparedStatement.getGeneratedKeys()) {
                    if (resultSet.next()) {
                        if (consumer != null) {
                            consumer.accept(resultSet.getLong(1));
                        }
                    }
                }
            }
        }
    }

    private ResultSet readNextFromQueue(Connection connection, String table, String channel) throws SQLException {
        PreparedStatement preparedStatement = connection.prepareStatement(flavor().readNextFromQueue(table));
        preparedStatement.setString(1, channel);
        preparedStatement.setTimestamp(2, Timestamp.valueOf(LocalDateTime.now()));
        return preparedStatement.executeQuery();
    }

    private void succeedInQueue(Connection connection, String table, Long key) throws SQLException {
        try (PreparedStatement preparedStatement = connection.prepareStatement(flavor().succeedInQueue(table))) {
            preparedStatement.setLong(1, key);
            preparedStatement.executeUpdate();
        }
    }

    private void failInQueue(Connection connection, String table, Long key) throws SQLException {
        try (PreparedStatement preparedStatement = connection.prepareStatement(flavor().succeedInQueue(table))) {
            preparedStatement.setLong(1, key);
            preparedStatement.executeUpdate();
        }
    }
}
